/*
 * array.c
 *
 * Author: Fredrik Noring   <noring@lysator.liu.se>
 *
 * (C) 1997 Fredrik Noring
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */

#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "array.h"

void *xmalloc(int size)
{
    void *p;
    
    p = malloc(size);
    if (!p) {
	perror("couldn't allocate memory");
	exit(1);
    }
    return p;
}

int element_size(struct array *a, int i)
{
    int n, j;

    if (i == a->n-1)
	    n = a->c[i+1] - a->c[i];
    else
	    n = a->c[i+1]-1 - a->c[i];
    if (n < 0) n = 0;

    for (j = 0; j < n; j++)
	    if (a->c[i][j] != '\r')
		    return n;
    return 0;
}

int get_hexint(struct array *a, int e, int r)
{
    char *s;
    int len, n = 0;

    s = a->c[e];
    len = element_size(a, e);
    while (len--) {
	if (*s >= '0' && *s <= '9')
	    n = (n << 4) + *s - '0';
	else if (*s >= 'a' && *s <= 'f')
	    n = (n << 4) + 10 + *s - 'a';
	else if (*s >= 'A' && *s <= 'F')
	    n = (n << 4) + 10 + *s - 'A';
	else if (*s == '\r');
	else {
		fprintf(stderr, "Strange number (line %d)\n", r+1);
		exit(1);
	}
	s++;
    }
    return n;
}

struct array *divide(char c, char *buf, int len)
{
    int n, i;
    char *b, **p;
    struct array *a;

    /* calculate number of characters */
    n = 1;
    b = buf;
    for (i = 0; i < len; i++)
	if (*(b++) == c)
	    n++;

    /* malloc array offsets */
    a = xmalloc(sizeof(struct array) + n*sizeof(char *));
    a->n = n;

    /* find all characters again, thus make the division */
    b = buf;
    p = a->c;
    *p++ = b;
    for (i = 0; i < len; i++)
	if (*(b++) == c)
	    *p++ = b;
    *p++ = b;

    return a;
}

int count_entries(struct array *rows, int r)
{
    int e = 0, c;
    struct array *cols;

    for ( ; r < rows->n; r++) {
	if (rows->c[r][0] == '_')
	    return e;

	cols = divide(' ', rows->c[r], element_size(rows, r));

	for (c = 0; c < cols->n; c++) {
		if (element_size(cols, c))
			e++;
	}
	free(cols);
    }
    return e;
}
