/*
 * lod2c version 0.5.0
 * DSP 56001 load format to C syntax converter
 *
 * Author: Fredrik Noring   <noring@lysator.liu.se>
 *
 * Written January 1997
 *
 * (C) 1997 Fredrik Noring
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>

#include "array.h"

int filesize(int fd)
{
	struct stat buf;
	fstat(fd, &buf);
	return buf.st_size;
}

void write_first(int fd, char *s)
{
	char s1[] = "char ";
	char s2[] = "[] = {\n";

	write(fd, s1, strlen(s1));
	write(fd, s, strlen(s));
	write(fd, s2, strlen(s2));
}

void write_number(int fd, int n)
{
	static int cols = 0;
	char s[256];

	if (cols) write(fd, ", ", 2);
	if (cols++ >= 4) {
		write(fd, "\n", 1);
		cols = 1;
	}
	sprintf(s, "0x%02x, 0x%02x, 0x%02x", ((char*) &n)[1] & 0xff,
		                             ((char*) &n)[2] & 0xff,
	                                     ((char*) &n)[3] & 0xff);
	write(fd, s, strlen(s));
}

void write_ending(int fd, char *s)
{
	char s1[] = " };\nint sizeof_";
	char s2[] = " = sizeof(";
	char s3[] = ");\n";

	write(fd, s1, strlen(s1));
	write(fd, s, strlen(s));
	write(fd, s2, strlen(s2));
	write(fd, s, strlen(s));
	write(fd, s3, strlen(s3));
}

int write_data(int fd, struct array *rows, int r, int addr, int mem)
{
    int offset = 0, n = 0, data, entries;

    /* Count number of _DATA entries */
    entries = count_entries(rows, r);

    /* Check if there are any entries */
    if (!entries) {
	    fprintf(stderr, "No data entries (line %d)\n", r+1);
	    exit(1);
    }

    /* Check that L data always are in pairs */
    if ((entries & 10) && (entries == 4 || entries == 5)) {
	    fprintf(stderr, "Uneven number of L entries (line %d)\n", r+1);
	    exit(1);
    }

    if (mem == 4) {
	mem = 1;
	offset = 1;
	entries /= 2;
    } else if (mem == 5) {
	mem = 2;
	offset = -1;
	entries /= 2;
    }

    write_number(fd, mem);
    write_number(fd, addr);
    write_number(fd, entries);

    while (r < rows->n) {
	int c;
	struct array *cols;

	if (n >= entries) return n;

	cols = divide(' ', rows->c[r], element_size(rows, r));
	for (c = 0; c < cols->n; c++) {
		if (element_size(cols, c)) {
			if (offset >= 0) {
				data = get_hexint(cols, c, r);
				write_number(fd, data);
				n++;
			}
			offset = -offset;
		}
	}
	free(cols);
	r++;
    }
    return n;
}

int main(int argc, char *argv[])
{
	int fd;
	int fdc = 1;
	char *s;

	if (argc < 2) {
		fprintf(stderr, "lod2c: <lod file> [output file]\n");
		return 0;
	}

	if (argc >= 3) {
		s = xmalloc(strlen(argv[2])+1);
		strcpy(s, argv[2]);
		if (strstr(s, "."))
			*strstr(s, ".") = '\0';
		fdc = open(argv[2], O_WRONLY | O_CREAT | O_TRUNC, 0644);
		if (fdc == -1) {
			perror("Couldn't open output file");
			exit(1);
		}
	} else if (argc == 2) {
		s = xmalloc(strlen(argv[1])+1);
		strcpy(s, argv[1]);
		if (strstr(s, "."))
			*strstr(s, ".") = '\0';
	} else {
		s = "output";
	}

  /* open lod file */
  fd = open(argv[1], O_RDONLY);
  if (fd == -1) {
	  perror("Couldn't open lod file");
  } else {
      int len;
      char *lod;

      len = filesize(fd);
      lod = mmap(0, len, PROT_READ, MAP_FILE | MAP_PRIVATE, fd, 0);
      if ((int) lod == -1) {
	  perror("couldn't mmap lod file");
      } else {
	  int r;
	  struct array *rows;

	  write_first(fdc, s);
	  
	  /* Loop for every newline */
	  rows = divide('\n', lod, len);
	  for (r = 0; r < rows->n; r++) {
	      struct array *cols;

	      cols = divide(' ', rows->c[r], element_size(rows, r));

	      /* Found _DATA record! */
	      if (strncmp(cols->c[0], "_DATA", 5) == 0) {
		  int addr;

		  /* Check if there are any rows below this record */
		  if (r+1 >= rows->n) {
		      fprintf(stderr, "No data entries (line %d)\n", r+1);
		      exit(1);
		  }

		  /* Check if there are enough parameters */
		  if (cols->n < 3) {
		      fprintf(stderr, "Data record not specified "
			      "(line %d)\n", r+1);
		      exit(1);
		  }

		  /* Parse _DATA address */
		  addr = get_hexint(cols, 2, r);

		  switch (*cols->c[1]) {
		  case 'L':
		      write_data(fdc, rows, r+1, addr, 4);  // X
		      write_data(fdc, rows, r+1, addr, 5);  // Y
		      break;
		  case 'Y':
		      write_data(fdc, rows, r+1, addr, 2);
		      break;
		  case 'X':
		      write_data(fdc, rows, r+1, addr, 1);
		      break;
		  case 'P':
		      write_data(fdc, rows, r+1, addr, 0);
		      break;
		  default:
		      fprintf(stderr, "Data record strange memory "
			      "(line %d)\n", r+1);
		      exit(1);
		  }
	      }
	      free(cols);
	  }
	  free(rows);

	  write_ending(fdc, s);
      }
      close(fd);
  }
  close(fdc);
  free(s);

  return 0;
}
